(define-module (uniseg graphemes stream)
  #:use-module (uniseg)
  #:use-module (uniseg internal)
  #:use-module (ice-9 textual-ports)
  #:use-module (ice-9 match)
  #:use-module (srfi srfi-71)
  #:use-module (srfi srfi-41)
  #:use-module (srfi srfi-9 gnu)
  #:export (make-grapheme
            grapheme?
            grapheme-glyphs
            grapheme-string
            grapheme-width
            string->grapheme-stream
            input->grapheme-stream))

;; TODO: the golang uniseg also does word and sentence boundaries. These state machines could be implemented if we wanted to.
(define-immutable-record-type <grapheme>
  (make-grapheme width glyphs-promise string-promise)
  grapheme?
  (width grapheme-width)
  (glyphs-promise _grapheme-glyphs-promise)
  (string-promise _grapheme-string-promise))

(define (grapheme-glyphs grapheme)
  "Return a lazily-constructed list of glyphs in the grapheme"
  (force (_grapheme-glyphs-promise grapheme)))

(define (grapheme-string grapheme)
  "Return a lazily-constructed string of the glyphs in the grapheme."
  (force (_grapheme-string-promise grapheme)))

(define (string->grapheme-stream str)
  "Given a string, create a (lazy) stream of graphemes."
  (with-input-from-string str
    (λ () (input->grapheme-stream (current-input-port)))))

(define-stream (input->grapheme-stream port)
  "Given an input port, create a (lazy) stream of graphemes."
  ;; The first is what state we are in, and the next is the grapheme
  ;; property of the current character.
  ;; Port of https://github.com/rivo/uniseg/blob/master/graphemerules.go
  (define (state-machine cur-state cur-prop)
    (match (list cur-state cur-prop)
      ;; Specifics need to go first, and then non-specifics afterwards
      ;; SPECIFIC RULES - no `_' in the match

      ;; Grapheme boundary #3s
      (('carriage-return 'line-feed)
       (values 'control+line-feed #f))

      ;; Grapheme boundary #6s
      (('hangul-syllable-l 'hangul-syllable-l)
       (values 'hangul-syllable-l #f))
      ((or ('hangul-syllable-l 'hangul-syllable-v)
           ('hangul-syllable-l 'hangul-syllable-lv))
       (values 'hangul-syllable-lv #f))
      (('hangul-syllable-l 'hangul-syllable-lvt)
       (values 'hangul-syllable-lvt #f))

      ;; Grapheme boundary #7s
      (('hangul-syllable-lv 'hangul-syllable-v)
       (values 'hangul-syllable-lv #f))
      (('hangul-syllable-lv 'hangul-syllable-t)
       (values 'hangul-syllable-lvt #f))

      ;; Grapheme boundary #8s
      (('hangul-syllable-lvt 'hangul-syllable-t)
       (values 'hangul-syllable-lvt #f))

      ;; Grapheme boundary #11s (emoji!)
      (('extended-pictographic 'extend)
       (values 'extended-pictographic #f))
      (('extended-pictographic 'zero-width-joiner)
       (values 'extended-pictographic+zero-width-joiner #f))
      (('extended-pictographic+zero-width-joiner 'extended-pictographic)
       (values 'extended-pictographic #f))

      ;; Grapheme boundaries #12s and #13s
      (('regional-indicator-odd 'regional-indicator)
       (values 'regioinal-indicator-even #f))
      (('regional-indicator-even 'regional-indicator)
       (values 'regional-indicator-odd #t))

      ;; NON-SPECIFIC RULES

      ;; Grapheme boundary #4n
      ((or ('carriage-return _)
           ('control+line-feed _))
       (values 'any #t))

      ;; Grapheme boundary #5n
      ((_ 'carriage-return)
       (values 'carriage-return #t))
      ((or (_ 'line-feed)
           (_ 'control))
       (values 'control+line-feed #t))
      ((_ 'hangul-syllable-l)
       (values 'hangul-syllable-l #t))

      ;; Grapheme boundary #7n
      ((or (_ 'hangul-syllable-lv)
           (_ 'hangul-syllable-v))
       (values 'hangul-syllable-lv #t))

      ;; Grapheme boundary #8n
      ((or (_ 'hangul-syllable-lvt)
           (_ 'hangul-syllable-t))
       (values 'hangul-syllable-lvt #t))

      ;; Grapheme boundary #9n
      ((or (_ 'extend)
           (_ 'zero-width-joiner))
       (values 'any #f))

      ;; Grapheme boundary #9n-A
      ((_ 'spacing-mark)
       (values 'any #f))

      ;; Grapheme boundary #9n-B
      (('prepend _)
       (values 'any #f))
      ((_ 'prepend)
       (values 'prepend #t))

      ;; Grapheme boundary #11n (emoji!)
      ((_ 'extended-pictographic)
       (values 'extended-pictographic #t))

      ;; Grapheme boundaries #12n and #13n
      ((_ 'regional-indicator)
       (values 'regional-indicator-odd #t))

      ;; Everything else considered a boundary
      (else (values 'any #t))))

  ;; State variables, initialized with the first glyph we get!
  (define first-glyph (get-char port))
  (define hit-eof? (eof-object? first-glyph))
  (define %glyphs-reverse (list first-glyph))
  (define %grapheme-width 0)

  (define (iterate-through-grapheme state)
    (define glyph (peek-char port))
    (if (eof-object? glyph)
        (begin
          (set! hit-eof? #t)
          state)
        (begin
          (let* ((width property (char-width glyph))
                 (next-state boundary? (state-machine state property)))

            (if boundary?
                state
                (let ((new-width (+ %grapheme-width width))
                      (new-glyphs-reverse (cons glyph %glyphs-reverse)))

                  ;; Officially induct this char into the cluster
                  (get-char port)
                  (set! %grapheme-width new-width)
                  (set! %glyphs-reverse new-glyphs-reverse)
                  (iterate-through-grapheme next-state)))))))

  ;; Need to explicitly get the first state
  ;; Skip all this if it's an empty stream at start
  (if hit-eof?
      stream-null
      (let ((first-width first-prop (char-width first-glyph)))
        (set! %grapheme-width first-width)
        (iterate-through-grapheme (state-machine 'any first-prop))

        (let ((grapheme (make-grapheme
                         %grapheme-width
                         ;; Delay to avoid construction of unnecessary lists and strings!
                         (delay (reverse %glyphs-reverse))
                         (delay (reverse-list->string %glyphs-reverse)))))

          ;; If we hit the eof here, we need one last stream entry, otherwise iterate further
          (stream-cons grapheme
                       (if hit-eof?
                           stream-null
                           (input->grapheme-stream port)))))))
